# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2024 SerialLab Corp.  All rights reserved.

from django.db import models
from django.utils.translation import gettext_lazy as _

from quartet_epcis.models import abstractmodels, events
from quartet_epcis.models.choices import EVENT_TYPE_CHOICES, ACTION_CHOICES


class Entry(abstractmodels.UUIDModel):
    """
    Represents an entry in the general ledger of serialized items and/or
    logical numbers used for serialized goods processing.
    """

    identifier = models.CharField(
        max_length=150,
        null=False,
        help_text=_("The primary unique id for the entry."),
        verbose_name=_("EPC URN"),
        db_index=True,
        unique=True,
    )
    parent_id = models.ForeignKey(
        "self",
        on_delete=models.SET_NULL,
        verbose_name=_("Parent ID"),
        related_name="parent_identifier",
        help_text=_("The parent of this identifier (if any)."),
        null=True,
    )
    top_id = models.ForeignKey(
        "self",
        on_delete=models.SET_NULL,
        verbose_name=_("Top ID"),
        help_text=_("The top level id (if any)."),
        related_name="top_identifier",
        null=True,
    )
    last_event = models.ForeignKey(
        "quartet_epcis.Event",
        null=True,
        on_delete=models.PROTECT,
        verbose_name=_("Last Event"),
        help_text=_("The last event to affect the status of this entry."),
    )
    last_event_time = models.DateTimeField(
        verbose_name=_("Last Event Time"),
        help_text=_(
            "The time of the event that last affected the status of" "this entry."
        ),
        null=True,
    )
    last_disposition = models.CharField(
        max_length=150,
        null=True,
        help_text=_(
            "The business condition of the objects associated "
            "with the EPCs, presumed to hold true until "
            "contradicted by a subsequent event.."
        ),
        verbose_name=_("Last Disposition"),
    )
    last_aggregation_event = models.ForeignKey(
        "quartet_epcis.Event",
        related_name="last_agg_event",
        null=True,
        on_delete=models.SET_NULL,
        verbose_name=_("Last Aggregation Event"),
        help_text=_(
            "Used mostly for internal tracking and performance during"
            "parsing.  This tracks the last aggregation event that"
            "affected the entry."
        ),
    )
    last_aggregation_event_time = models.DateTimeField(
        verbose_name=_("Last Aggregation Event Time"),
        help_text=_(
            "The time of the event that last affected the status of"
            "this entries hierarchical relation to other entries."
        ),
        null=True,
    )
    last_aggregation_event_action = models.CharField(
        null=True,
        max_length=10,
        verbose_name=_("Last Aggregation Action"),
        help_text=_(
            "The action (ADD or DELETE) of the last aggregation "
            "event that affected this entry.  Observation events "
            "are not noted."
        ),
        choices=ACTION_CHOICES,
    )
    is_parent = models.BooleanField(
        default=False,
        null=False,
        verbose_name=_("Is Parent"),
        help_text=_(
            "True if this entry is a parent in any hierarchies. False" "if not."
        ),
    )
    decommissioned = models.BooleanField(
        default=False,
        null=False,
        verbose_name=_("Decommissioned"),
        help_text=_(
            "Whether or not the entry has been decommissioned.  Once"
            "an entry is decommissioned, it can no longer take place"
            "in business processes."
        ),
    )

    @property
    def is_top(self):
        return self.is_parent and self.top_id is None

    def __str__(self):
        return self.identifier or ""

    class Meta:
        verbose_name = _("Entry")
        verbose_name_plural = _("Entries")
        app_label = "quartet_epcis"
        ordering = ["created"]


class EntryEvent(models.Model):
    """
    An intersection entity for events and entries.
    """

    event = models.ForeignKey(
        "quartet_epcis.Event",
        null=False,
        help_text=_("The UUID of the event."),
        verbose_name=_("Event ID"),
        db_index=True,
        on_delete=models.CASCADE,
    )
    event_type = models.CharField(
        null=False,
        max_length=3,
        verbose_name=_("Event Type"),
        help_text=_(
            "The type of event (Aggregation, Object, Transaction or" " Transformation."
        ),
        choices=EVENT_TYPE_CHOICES,
    )
    event_time = models.DateTimeField(
        null=False,
        help_text=_("The Event's eventTime."),
        verbose_name=_("Event Time"),
        db_index=True,
    )
    entry = models.ForeignKey(
        "quartet_epcis.Entry",
        null=False,
        help_text=_("The Unique ID of the Entry"),
        verbose_name=_("Entry ID"),
        db_index=True,
        on_delete=models.CASCADE,
    )
    identifier = models.CharField(
        max_length=150,
        null=False,
        help_text=_("A redundant entry ID entry for fast event composition."),
        verbose_name=_("EPC URN"),
        db_index=True,
    )
    is_parent = models.BooleanField(
        default=False,
        help_text=_(
            "Whether or not this entry was the parent of it's " "constituent event."
        ),
        verbose_name=_("Is Event Parent"),
    )
    output = models.BooleanField(
        default=False,
        help_text=_(
            "Whether or not the entry was the output of " "a Transformation event."
        ),
        verbose_name=_("Transformation Output"),
    )
    created = models.DateTimeField(
        auto_now_add=True,
        verbose_name=_("Created"),
        help_text=_("When this record was created."),
    )
    modified = models.DateTimeField(
        auto_now=True,
        verbose_name=_("Modified"),
        help_text=_("When this record was last modified."),
    )
    task_name = models.CharField(
        max_length=150,
        null=True,
        blank=True,
        verbose_name=_("Task Name"),
        help_text=_("The name of the Task that parsed this Entry and Event."),
    )

    def __str__(self):
        return self.identifier or ""

    class Meta:
        verbose_name = _("Entry Event Record")
        verbose_name_plural = _("Entry Event Records")
        models.indexes = [models.Index(fields=["event", "entry"])]
        app_label = "quartet_epcis"



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2024 SerialLab Corp.  All rights reserved.


from rest_framework.routers import DefaultRouter
from quartet_epcis import viewsets


router = DefaultRouter()
router.register(r"entries", viewsets.EntryViewSet, basename="entries")
router.register(r"entry-events", viewsets.EntryEventViewSet, basename="entry-events")
router.register(r"events", viewsets.EventViewSet, basename="events")
router.register(
    r"transformation-ids",
    viewsets.TransformationIDViewSet,
    basename="transformation-ids",
)
router.register(
    r"error-declarations",
    viewsets.ErrorDeclarationViewSet,
    basename="error-declarations",
)
router.register(
    r"quantity-elements", viewsets.QuantityElementViewSet, basename="quantity-elements"
)
router.register(
    r"business-transactions",
    viewsets.BusinessTransactionViewSet,
    basename="business-transactions",
)
router.register(r"ilmd", viewsets.InstanceLotMasterDataViewSet, basename="ilmd")
router.register(r"sources", viewsets.SourceViewSet, basename="sources")
router.register(r"source-events", viewsets.SourceEventViewSet, basename="source-events")
router.register(r"destinations", viewsets.DestinationViewSet, basename="destinations")
router.register(
    r"destination-event", viewsets.DestinationEventViewSet, basename="destinations"
)
router.register(r"messages", viewsets.MessageViewSet, basename="messages")



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2024 SerialLab Corp.  All rights reserved.

from django.db.models import Q
from django.utils.translation import gettext as _
from django.contrib.auth.models import Group, Permission
from django.core.management.base import BaseCommand
from quartet_epcis.models import entries, events, headers


class Command(BaseCommand):
    help = _("Creates the default EPCIS access groups.")

    def handle(self, *args, **options):
        print("Creating the EPCIS Access group...")
        group, created = Group.objects.get_or_create(name="EPCIS Access")
        if created:
            permissions = Permission.objects.filter(
                Q(codename__endswith="_entry")
                | Q(codename__endswith="_entryevent")
                | Q(codename__endswith="_errordeclaration")
                | Q(codename__endswith="_event")
                | Q(codename__endswith="_documentidentification")
                | Q(codename__endswith="_businesstransaction")
                | Q(codename__endswith="_destination")
                | Q(codename__endswith="_source")
                | Q(codename__endswith="_instancelotmasterdata")
                | Q(codename__endswith="_message")
                | Q(codename__endswith="_partner")
                | Q(codename__endswith="_quantityelement")
                | Q(codename__endswith="_sbdh")
                | Q(codename__endswith="_sourceevent")
                | Q(codename__endswith="_transformationid")
                | Q(codename__endswith="_destinationevent")
            )
            group.permissions.set(permissions)
            print("EPCIS Access group created.")
        else:
            print('Group "EPCIS Access" already exists.')



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2024 SerialLab Corp.  All rights reserved.
from django.conf import settings
from django.core.management.base import BaseCommand
from django.utils.translation import gettext as _

from quartet_epcis.models import entries, events, headers


class Command(BaseCommand):
    help = _("Clears out the event data.  For testing and debug systems only.")

    def add_arguments(self, parser):
        parser.add_argument(
            "--force", dest="force", help="Force the removal of the event data."
        )

    def handle(self, *args, **options):
        debug = getattr(settings, "DEBUG", False)
        force = options.get("force", False)
        if debug or force in ["True", "true", "TRUE"]:
            print("Clearing out event data.")
            entries.Entry.objects.all().delete()
            events.Event.objects.all().delete()
            entries.EntryEvent.objects.all().delete()
            headers.Message.objects.all().delete()
            headers.SBDH.objects.all().delete()
            headers.Partner.objects.all().delete()
            headers.DocumentIdentification.objects.all().delete()
            print("Done.")
        else:
            print("Not valid on non-debug systems.")



# -*- coding: utf-8 -*-
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2024 SerialLab Corp.  All rights reserved.
from django.urls import re_path
from quartet_epcis.routers import router
from quartet_epcis import views

app_name = "quartet_epcis"

urlpatterns = [
    re_path(r"^event-detail/?$", views.EventDetailView.as_view(), name="event-detail"),
    re_path(
        r"^event-detail/(?P<event_id>[0-9a-fA-F\-_]{1,50})/$",
        views.EventDetailView.as_view(),
        name="event-detail",
    ),
    re_path(
        r"^events-by-entry-id/?$",
        views.EntryEventHistoryView.as_view(),
        name="events-by-entry-id",
    ),
    re_path(
        r"^events-by-entry-id/(?P<entry_identifier>[0-9a-zA-Z\W]{1,100})/$",
        views.EntryEventHistoryView.as_view(),
        name="events-by-entry-id",
    ),
    re_path(
        r"^events-by-entry-pk/?$",
        views.EntryEventHistoryView.as_view(),
        name="events-by-entry-pk",
    ),
    re_path(
        r"^events-by-entry-pk/(?P<entry_pk>[0-9a-zA-Z\W]{1,100})/$",
        views.EntryEventHistoryView.as_view(),
        name="events-by-entry-pk",
    ),
    re_path(
        r"^events-by-ilmd/?$", views.EventsByILMDView.as_view(), name="events-by-ilmd"
    ),
    re_path(
        r"^events-by-ilmd/(?P<ilmd_name>[[\w\s\W]{1,150})/"
        r"(?P<ilmd_value>[[\w\s\W]"
        r"{1,255})/$",
        views.EventsByILMDView.as_view(),
        name="events-by-ilmd",
    ),
    re_path(r"^message/?$", views.MessageDetail.as_view(), name="message"),
    re_path(
        r"^message/(?P<message_id>[\w-]{1,50})/$",
        views.MessageDetail.as_view(),
        name="message",
    ),
]

urlpatterns += router.urls



# Generated by Django 4.2.4 on 2023-09-04 16:03

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("quartet_epcis", "0005_auto_20200902_1023"),
    ]

    operations = [
        migrations.AlterField(
            model_name="businesstransaction",
            name="id",
            field=models.BigAutoField(
                auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
            ),
        ),
        migrations.AlterField(
            model_name="destinationevent",
            name="id",
            field=models.BigAutoField(
                auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
            ),
        ),
        migrations.AlterField(
            model_name="documentidentification",
            name="id",
            field=models.BigAutoField(
                auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
            ),
        ),
        migrations.AlterField(
            model_name="entryevent",
            name="id",
            field=models.BigAutoField(
                auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
            ),
        ),
        migrations.AlterField(
            model_name="errordeclaration",
            name="id",
            field=models.BigAutoField(
                auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
            ),
        ),
        migrations.AlterField(
            model_name="instancelotmasterdata",
            name="id",
            field=models.BigAutoField(
                auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
            ),
        ),
        migrations.AlterField(
            model_name="message",
            name="id",
            field=models.BigAutoField(
                auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
            ),
        ),
        migrations.AlterField(
            model_name="partner",
            name="id",
            field=models.BigAutoField(
                auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
            ),
        ),
        migrations.AlterField(
            model_name="quantityelement",
            name="id",
            field=models.BigAutoField(
                auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
            ),
        ),
        migrations.AlterField(
            model_name="sbdh",
            name="id",
            field=models.BigAutoField(
                auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
            ),
        ),
        migrations.AlterField(
            model_name="sourceevent",
            name="id",
            field=models.BigAutoField(
                auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
            ),
        ),
    ]



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2024 SerialLab Corp.  All rights reserved.

from threading import Lock
from django.db import models
from django.utils.translation import gettext_lazy as _
from model_utils import models as utils
from haikunator import Haikunator


haiku = Haikunator()


def haikunate():
    """
    Since the haikunator is a class method
    it could not be used directly as a default callable for
    a django field...hence this function.
    """
    try:
        lock = Lock()
        lock.acquire()
        ret = haiku.haikunate(token_length=8, token_hex=True)
    finally:
        lock.release()
    return ret


document_type_choices = (
    ("Events", "Events"),
    ("MasterData", "MasterData"),
    ("QueryControl-Request", "QueryControl-Request"),
    ("QueryControl-Response", "QueryControl-Response"),
    ("QueryCallback", "QueryCallback"),
    ("Query", "Query"),
)

partner_choices = (("Sender", "Sender"), ("Receiver", "Receiver"))


class Message(models.Model):
    """
    An umbrella model to unify all headers and events that arrived
    as part of a single message.  If the message had a unique id supplied in
    the header, that will be used (albeit redunantly) along with a UUID4
    for each message as the primary key.
    """

    utils.AutoCreatedField(
        help_text=_("The date and time this record was created."),
        verbose_name=_("Created Date"),
    )

    def __str__(self):
        return str(self.id) or ""

    class Meta:
        app_label = "quartet_epcis"


class SBDH(models.Model):
    """
    The db model representing the sbdh as defined by GS1.
    """

    header_version = models.CharField(
        max_length=10,
        null=False,
        help_text=_(
            "Descriptor which contains version "
            "information for the SBDH.  Default is 1.0"
        ),
        verbose_name=_("Header Version"),
        default="1.0",
    )
    document_identification = models.ForeignKey(
        "quartet_epcis.DocumentIdentification", null=False, on_delete=models.CASCADE
    )
    message = models.ForeignKey(
        "quartet_epcis.Message",
        on_delete=models.CASCADE,
        help_text=_("The message this header was associated with."),
        verbose_name=_("Message"),
    )

    def __str__(self):
        return "%s: %s" % (self.document_identification, self.message) or ""

    class Meta:
        verbose_name = _("SBDH")
        verbose_name_plural = _("SBDHs")
        app_label = "quartet_epcis"


class Partner(models.Model):
    """
    Trading partner data.
    """

    partner_type = models.CharField(
        max_length=20,
        null=False,
        help_text=_("The type of partner.  Either Sender or Receiver."),
        verbose_name=_("Partner Type"),
        choices=partner_choices,
    )
    authority = models.CharField(
        max_length=20,
        null=True,
        help_text=_(
            "The identifying authority/format for the "
            "identification field. Default is SGLN."
        ),
        verbose_name=_("authority"),
        default="SGLN",
    )
    identifier = models.CharField(
        max_length=100,
        null=True,
        help_text=_(
            "An identifier that is in line with the authority "
            "specified in the authority field.  Typically an "
            "SGLN URN value."
        ),
        verbose_name=_("Identifier"),
    )
    contact = models.CharField(
        max_length=50,
        null=True,
        help_text=_("The contact/name info."),
        verbose_name=_("Contact"),
    )
    email_address = models.EmailField(
        max_length=100,
        null=True,
        help_text=_("The email address for the partner."),
        verbose_name=_("Email Address"),
    )

    fax_number = models.CharField(
        max_length=20,
        null=True,
        help_text=_("Fax number."),
        verbose_name=_("Fax Number"),
    )
    telephone_number = models.CharField(
        max_length=20,
        null=True,
        help_text=_("Telephone number for the partner."),
        verbose_name=_("Telephone number."),
    )
    contact_type_identifier = models.CharField(
        max_length=40,
        null=True,
        help_text=_("The type of contact- for example, "),
        verbose_name=_("Role of the contact in this business process."),
    )
    header = models.ForeignKey(
        "quartet_epcis.SBDH",
        on_delete=models.CASCADE,
        help_text=_("The related SBDH."),
        verbose_name=_("SBDH"),
    )

    def __str__(self):
        if self.authority:
            ret = "{0}:{1}".format(self.authority, self.identifier) or ""
        else:
            ret = "{0}:{1}".format(self.id, self.partner_type) or ""
        return ret

    class Meta:
        verbose_name = _("Partner")
        verbose_name_plural = _("Partners")
        app_label = "quartet_epcis"


class DocumentIdentification(models.Model):
    """
    The core model for the GS1 SBDH document identification element.
    """

    standard = models.CharField(
        max_length=20,
        null=False,
        default="EPCglobal",
        help_text=_(
            "The originator of the standard that the "
            "following document falls under.  "
            "Default is EPCglobal."
        ),
        verbose_name=_("Standard"),
    )
    type_version = models.CharField(
        max_length=10,
        null=False,
        help_text=_(
            "Descriptor which contains versioning information "
            "or number of the standard that defines the document"
            "which is specified in the ’Type’ data element, e.g. "
            "values could be ‘1.3’ or ‘D.96A’, etc."
        ),
        verbose_name=_("Version"),
        default="1.0",
    )
    instance_identifier = models.CharField(
        max_length=100,
        null=False,
        help_text=_(
            "Descriptor which contains reference information "
            "which uniquely identifies this instance of the SBD "
            "between the sender and the receiver."
        ),
        verbose_name=_("Instance Identifier"),
        db_index=True,
    )
    document_type = models.CharField(
        max_length=25,
        null=False,
        help_text=_(
            "A logical indicator representing the type of "
            "Business Data being sent or the named type of "
            "business data."
        ),
        verbose_name=_("Type"),
        choices=document_type_choices,
    )
    multiple_type = models.BooleanField(
        null=True,
        default=False,
        help_text=_(
            "A flag to indicate that there is more than one "
            "type of Document in the instance."
        ),
        verbose_name=_("Multiple Type"),
    )
    creation_date_and_time = models.CharField(
        max_length=35,
        null=True,
        help_text=_(
            "Descriptor which contains date and time of " "SBDH/document creation."
        ),
        verbose_name=_("Creation Date and Time"),
    )

    def __str__(self):
        return "Document Id {0}: {1}".format(self.id, self.creation_date_and_time) or ""

    class Meta:
        verbose_name = _("Document Identification")
        verbose_name_plural = _("Document Identifications")
        app_label = "quartet_epcis"



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2024 SerialLab Corp.  All rights reserved.

import uuid
from threading import Lock
from django.db import models
from django.utils.translation import gettext_lazy as _
from django.utils import timezone
from quartet_epcis.models import choices
from haikunator import Haikunator
from EPCPyYes.core.v1_2 import CBV

haiku = Haikunator()


def haikunate():
    """
    Since the haikunator is a class method
    it could not be used directly as a default callable for
    a django field...hence this function.
    """
    try:
        lock = Lock()
        lock.acquire()
        ret = haiku.haikunate(token_length=8, token_hex=True)
    finally:
        lock.release()
    return ret


class UUIDModel(models.Model):
    """
    A base model which uses UUIDs for primary key values.
    """

    id = models.UUIDField(
        primary_key=True,
        default=uuid.uuid4,
        editable=False,
        help_text=_("Unique ID"),
        verbose_name=_("Unique ID"),
    )

    created = models.DateTimeField(
        auto_now_add=True,
        verbose_name=_("Created"),
        help_text=_("When this record was created."),
    )
    modified = models.DateTimeField(
        auto_now=True,
        verbose_name=_("Modified"),
        help_text=_("When this record was last modified."),
    )

    class Meta:
        abstract = True


class SourceModel(models.Model):
    """
    Abstract model class for any classes that reference a UUIDModel
    super-class.  Includes the UUID model UUID and a reference to the
    object type.
    """

    source_event_id = models.UUIDField(
        help_text=_("The UUID of the source event within the " "Quartet database."),
        verbose_name=_("Source Event UUID"),
        null=False,
    )
    source_event_type = models.CharField(
        max_length=2,
        choices=choices.EVENT_TYPE_CHOICES,
        help_text=_(
            "The type of originating event: Object, Aggregation, "
            "Transaction or Transformation."
        ),
        verbose_name=_("Source Event Type"),
    )

    def __str__(self):
        return str(self.source_event_id) or ""

    class Meta:
        abstract = True


class EPCISEvent(UUIDModel):
    """
    The base EPCIS event as defined by GS1 on page 38 of the EPCIS 1.2 draft.
    """

    event_time = models.DateTimeField(
        db_index=True,
        null=False,
        editable=False,
        help_text=_(
            "The date and time at which the EPCIS Capturing "
            "Application asserts the event occurred."
        ),
        verbose_name=_("Event Time"),
    )
    event_timezone_offset = models.CharField(
        max_length=6,
        null=True,
        help_text=_(
            "The time zone offset in effect at the "
            "time and place the event occurred, expressed as an "
            "offset from UTC"
        ),
        verbose_name=_("Event Timezone Offset"),
        default="+00:00",
    )
    record_time = models.DateTimeField(
        null=True,
        help_text=_(
            "The date and time at which this event was"
            " recorded by an EPCIS Repository."
        ),
        verbose_name=_("Record Time"),
        default=timezone.now,
    )
    event_id = models.CharField(
        max_length=150,
        null=True,
        default=uuid.uuid4,
        help_text=_(
            "An identifier for this event as specified by the "
            "capturing application, globally unique across all events "
            "other than error declarations. Not to be confused with "
            "the unique id/primary key for events within a database."
        ),
        verbose_name=_("Event ID"),
        db_index=True,
    )

    def __str__(self):
        return self.event_time.utcnow().isoformat() or ""

    class Meta:
        abstract = True


class EPCISBusinessEvent(EPCISEvent):
    """
    Abstract base-class for super-classes with an Action,
    biz step, biz location, etc...basically
    every main EPCIS class except the TransformationEvent class.
    """

    action = models.CharField(
        max_length=10,
        choices=choices.ACTION_CHOICES,
        null=False,
        help_text=_(
            "How this event relates to the lifecycle of the "
            "EPCs named in this event."
        ),
        verbose_name=_("Action"),
    )
    biz_step = models.CharField(
        max_length=150,
        null=True,
        help_text=_("The business step of which this event was a part."),
        verbose_name=_("Business Step"),
        choices=[(tag.value, tag.value) for tag in CBV.BusinessSteps],
        blank=True,
    )
    disposition = models.CharField(
        max_length=150,
        null=True,
        help_text=_(
            "The business condition of the objects associated "
            "with the EPCs, presumed to hold true until "
            "contradicted by a subsequent event.."
        ),
        verbose_name=_("Disposition"),
        choices=[(tag.value, tag.value) for tag in CBV.Disposition],
        blank=True,
    )
    read_point = models.CharField(
        max_length=150,
        null=True,
        help_text=_("The read point at which the event took place."),
        verbose_name=_("Read Point"),
        blank=True,
    )
    biz_location = models.CharField(
        max_length=150,
        null=True,
        help_text=_(
            "The business location where the objects "
            "associated with the EPCs may be found, "
            "until contradicted "
            "by a subsequent event."
        ),
        verbose_name=_("Business Location"),
        blank=True,
    )

    def __str__(self):
        return self.event_time.utcnow().isoformat() or ""

    class Meta:
        abstract = True
